"""
The Agent class is where all the magic happens!

Inheriting the Agent class allows you to modify the behaviour of the agents in the simulation.
"""

from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING, Any, Callable, Generator, Optional, Union

import pygame as pg
from pygame.mask import Mask
from pygame.math import Vector2
from pygame.rect import Rect
from pygame.sprite import Group, Sprite
from pygame.surface import Surface
from typing_extensions import Self

from .config import Config
from .proximity import ProximityIter
from .util import random_angle, random_pos, round_pos

if TYPE_CHECKING:
    from .obstacle import Obstacle
    from .simulation import HeadlessSimulation, Shared


class Agent(Sprite):
    """
    The `Agent` class is home to Violet's various additions and is
    built on top of [PyGame's Sprite](https://www.pygame.org/docs/ref/sprite.html) class.

    While you can simply add this `Agent` class to your simulations by calling `batch_spawn_agents`,
    the agents won't actually do anything interesting.
    Sure, they'll move around the screen very energetically, but they don't *interact* with each other.

    That's where you come in!
    By inheriting this `Agent` class you can make use of all its utilities,
    while also programming the behaviour of your custom agent.
    """

    id: int
    """The unique identifier of the agent."""

    config: Config
    """The config of the simulation that's shared with all agents.
    
    The config can be overriden when inheriting the Agent class.
    However, the config must always:

    1. Inherit `Config`
    2. Be decorated by `@deserialize` and `@dataclass`
    """

    shared: Shared
    """Attributes that are shared between the simulation and all agents."""

    _images: list[Surface]
    """A list of images which you can use to change the current image of the agent."""

    _image_index: int
    """The currently selected image."""

    _image_cache: Optional[tuple[int, Surface]] = None

    _area: Rect
    """The area in which the agent is free to move."""

    move: Vector2
    """The current angle and speed used for the agent's movement.
    
    This property is also used to automatically rotate the agent's image
    when `vi.config.Schema.image_rotation` is enabled.
    """

    pos: Vector2
    """The current (centre) position of the agent."""

    __previous_move: Optional[Vector2] = None
    """The value of the `move` vector before the agent was frozen."""

    _obstacles: Group
    """The group of obstacles the agent can collide with."""

    _sites: Group
    """The group of sites on which the agent can appear."""

    __simulation: HeadlessSimulation

    _still_stuck: bool = False

    def __init__(
        self,
        images: list[Surface],
        simulation: HeadlessSimulation,
        pos: Optional[Vector2] = None,
        move: Optional[Vector2] = None,
    ):
        Sprite.__init__(self, simulation._all, simulation._agents)

        self.__simulation = simulation

        self.id = simulation._agent_id()
        self.config = simulation.config
        self.shared = simulation.shared

        # Default to first image in case no image is given
        self._image_index = 0
        self._images = images

        self._obstacles = simulation._obstacles
        self._sites = simulation._sites

        self._area = simulation._area
        self.move = (
            move
            if move is not None
            else random_angle(self.config.movement_speed, prng=self.shared.prng_move)
        )

        # On spawn acts like the __init__ for non-pygame facing state.
        # It could be used to override the initial image if necessary.
        self.on_spawn()

        if pos is not None:
            self.pos = pos

        if not hasattr(self, "pos"):
            # Keep changing the position until the position no longer collides with any obstacle.
            while True:
                self.pos = random_pos(self._area, prng=self.shared.prng_move)

                obstacle_hit = pg.sprite.spritecollideany(self, self._obstacles, pg.sprite.collide_mask)  # type: ignore
                if not bool(obstacle_hit) and self._area.contains(self.rect):
                    break

    def _get_image(self) -> Surface:
        image = self._images[self._image_index]

        if self.config.image_rotation:
            angle = self.move.angle_to(Vector2((0, -1)))

            return pg.transform.rotate(image, angle)
        else:
            return image

    @property
    def image(self) -> Surface:
        """The read-only image that's used for PyGame's rendering."""

        if self._image_cache is not None:
            frame, image = self._image_cache
            if frame == self.shared.counter:
                return image

        new_image = self._get_image()
        self._image_cache = (self.shared.counter, new_image)

        return new_image

    @property
    def center(self) -> tuple[int, int]:
        """The read-only centre position of the agent."""

        return round_pos(self.pos)

    @property
    def rect(self) -> Rect:
        """The read-only bounding-box that's used for PyGame's rendering."""

        rect = self.image.get_rect()
        rect.center = self.center

        return rect

    @property
    def mask(self) -> Mask:
        """A read-only bit-mask of the image used for collision detection with obstacles and sites."""

        return pg.mask.from_surface(self.image)

    def update(self):
        """Run your own agent logic at every tick of the simulation.
        Every frame of the simulation, this method is called automatically for every agent of the simulation.

        To add your own logic, inherit the `Agent` class and override this method with your own.
        """

        ...

    def obstacle_intersections(
        self, scale: float = 1
    ) -> Generator[Vector2, None, None]:
        """Retrieve the centre coordinates of all obstacle intersections.

        If you not only want to check for obstacle collision,
        but also want to retrieve the coordinates of pixel groups
        that are colliding with your agent, then `obstacle_intersections` is for you!

        If you only have one obstacle in your environment,
        but it doesn't have a fill, only a stroke,
        then your agent could possibly be colliding with different areas of the stroke.
        Therefore, this method checks the bitmasks of both the agent and the obstacle
        to calculate the overlapping bitmask.
        From this overlapping bitmask,
        the centre coordinates of all groups of connected pixels are returned.

        To emulate a bigger (or smaller) radius,
        you can pass along the `scale` option.
        A scale of 2 makes your agent twice as big,
        but only for calculating the intersecting bitmasks.
        """

        mask = pg.mask.from_surface(self.image)

        # Scale the mask to the desired size
        width, height = mask.get_size()
        mask = mask.scale((width * scale, height * scale))

        # Align the mask to the centre position of the agent
        rect = mask.get_rect()
        rect.center = self.center

        for sprite in self._obstacles.sprites():
            obstacle: Obstacle = sprite  # type: ignore

            # Calculate the mask offset
            x = obstacle.rect.x - rect.x
            y = obstacle.rect.y - rect.y

            overlap = mask.overlap_mask(obstacle.mask, offset=(x, y))

            # For some reason PyGame has the wrong type hint here (single instead of list)
            overlap_rects: list[pg.rect.Rect] = overlap.get_bounding_rects()  # type: ignore

            for overlap_rect in overlap_rects:
                # Undo the offset
                overlap_rect.x += rect.x
                overlap_rect.y += rect.y

                # Return the centre coordinates of the connected pixels
                yield Vector2(overlap_rect.center)

    def on_spawn(self):
        """Run any code when the agent is spawned into the simulation.

        This method is a replacement for `__init__`, which you should not overwrite directly.
        Instead, you can make alterations to your Agent within this function instead.

        You should override this method when inheriting Agent to add your own logic.

        Some examples include:
        - Changing the image or state of your Agent depending on its assigned identifier.
        """

        ...

    def there_is_no_escape(self) -> bool:
        """Pac-Man-style teleport the agent to the other side of the screen when it is outside of the playable area.

        Examples
        --------

        An agent that will always move to the right, until snapped back to reality.

        >>> class MyAgent(Agent):
        ...     def on_spawn(self):
        ...         self.move = Vector2((5, 0))
        ...
        ...     def change_position(self):
        ...         self.there_is_no_escape()
        ...         self.pos += self.move
        """

        changed = False

        if self.pos.x < self._area.left:
            changed = True
            self.pos.x = self._area.right

        if self.pos.x > self._area.right:
            changed = True
            self.pos.x = self._area.left

        if self.pos.y < self._area.top:
            changed = True
            self.pos.y = self._area.bottom

        if self.pos.y > self._area.bottom:
            changed = True
            self.pos.y = self._area.top

        return changed

    def change_position(self):
        """Change the position of the agent.

        The agent's new position is calculated as follows:
        1. The agent checks whether it's outside of the visible screen area.
        If this is the case, then the agent will be teleported to the other edge of the screen.
        2. If the agent collides with any obstacles, then the agent will turn around 180 degrees.
        3. If the agent has not collided with any obstacles, it will have the opportunity to slightly change its angle.
        """
        changed = self.there_is_no_escape()

        prng = self.shared.prng_move

        # Always calculate the random angle so a seed could be used.
        deg = prng.uniform(-30, 30)

        # Only update angle if the agent was teleported to a different area of the simulation.
        if changed:
            self.move.rotate_ip(deg)

        # Obstacle Avoidance
        obstacle_hit = pg.sprite.spritecollideany(self, self._obstacles, pg.sprite.collide_mask)  # type: ignore
        collision = bool(obstacle_hit)

        # Reverse direction when colliding with an obstacle.
        if collision and not self._still_stuck:
            self.move.rotate_ip(180)
            self._still_stuck = True

        if not collision:
            self._still_stuck = False

        # Random opportunity to slightly change angle.
        # Probabilities are pre-computed so a seed could be used.
        should_change_angle = prng.random()
        deg = prng.uniform(-10, 10)

        # Only allow the angle opportunity to take place when no collisions have occured.
        # This is done so an agent always turns 180 degrees. Any small change in the number of degrees
        # allows the agent to possibly escape the obstacle.
        if not collision and not self._still_stuck and 0.25 > should_change_angle:
            self.move.rotate_ip(deg)

        # Actually update the position at last.
        self.pos += self.move

    def in_proximity_accuracy(self) -> ProximityIter[tuple[Self, float]]:
        """Retrieve other agents that are in the `vi.config.Schema.radius` of the current agent.

        This proximity method calculates the distances between agents to determine whether
        an agent is in the radius of the current agent.

        To calculate the distances between agents, up to four chunks have to be retrieved from the Proximity Engine.
        These extra lookups, in combination with the vector distance calculation, have a noticable impact on performance.
        Note however that this performance impact is only noticable with >1000 agents.

        If you want to speed up your simulation at the cost of some accuracy,
        consider using the `in_proximity_performance` method instead.

        This function doesn't return the agents as a `list` or as a `set`.
        Instead, you are given a `vi.proximity.ProximityIter`, a small wrapper around a Python generator.

        Examples
        --------

        Count the number of agents that are in proximity
        and change to image `1` if there is at least one agent nearby.

        >>> class MyAgent(Agent):
        ...     def update(self):
        ...         in_proximity = self.in_proximity_accuracy().count()
        ...
        ...         if in_proximity >= 1:
        ...             self.change_image(1)
        ...         else:
        ...             self.change_image(0)

        Kill the first `Human` agent that's in proximity.

        >>> class Zombie(Agent):
        ...     def update(self):
        ...         human = (
        ...             self.in_proximity_accuracy()
        ...             .without_distance()
        ...             .filter_kind(Human) # 👈 don't want to kill other zombies
        ...             .first() # 👈 can return None if no humans are around
        ...         )
        ...
        ...         if human is not None:
        ...             human.kill()

        Calculate the average distance of agents that are in proximity.

        >>> class Heimerdinger(Agent):
        ...     def update(self):
        ...         in_proximity = list(self.in_proximity_accuracy())
        ...
        ...         dist_sum = sum(dist for agent, dist in in_proximity)
        ...
        ...         dist_avg = (
        ...             dist_sum / len(in_proximity)
        ...             if len(in_proximity) > 0
        ...             else 0
        ...         )
        """

        return self.__simulation._proximity.in_proximity_accuracy(self)

    def in_proximity_performance(self) -> ProximityIter[Self]:
        """Retrieve other agents that are in the `vi.config.Schema.radius` of the current agent.

        Unlike `in_proximity_accuracy`, this proximity method does not calculate the distances between agents.
        Instead, it retrieves agents that are in the same chunk as the current agent,
        irrespective of their position within the chunk.

        If you find yourself limited by the performance of `in_proximity_accuracy`,
        you can swap the function call for this one instead.
        This performance method roughly doubles the frame rates of the simulation.
        """

        return self.__simulation._proximity.in_proximity_performance(self)

    def on_site(self) -> bool:
        """Check whether the agent is currently on a site.

        Examples
        --------

        Stop the agent's movement when it reaches a site (think of a nice beach).

        >>> class TravellingAgent(Agent):
        ...     def update(self):
        ...         if self.on_site():
        ...             # crave that star damage
        ...             self.freeze_movement()
        """

        return bool(
            pg.sprite.spritecollideany(self, self._sites, pg.sprite.collide_mask)  # type: ignore
        )

    def freeze_movement(self):
        """Freeze the movement of the agent. The movement can be continued by calling `continue_movement`."""

        self.__previous_move = self.move
        self.move = Vector2(0, 0)

    def continue_movement(self):
        """Continue the movement of the agent by using the angle and speed of the agent before its movement was frozen."""

        if self.__previous_move is not None:
            self.move = self.__previous_move
            self.__previous_move = None

    def change_image(self, index: int):
        """Change the image of the agent.

        If you want to change the agent's image to the second image in the images list,
        then you can change the image to index 1:
        >>> self.change_image(1)
        """

        self._image_index = index

    def save_data(self, column: str, value: Any):
        """Add extra data to the simulation's metrics.

        The following data is collected automatically:
        - agent identifier
        - current frame
        - x and y coordinates

        Examples
        --------

        Saving the number of agents that are currently in proximity:

        >>> class MyAgent(Agent):
        ...     def update(self):
        ...         in_proximity = self.in_proximity_accuracy().count()
        ...         self.save_data("in_proximity", in_proximity)
        """

        self.__simulation._metrics._temporary_snapshots[column].append(value)

    def _collect_replay_data(self):
        """Add the minimum data needed for the replay simulation to the dataframe."""

        x, y = self.center
        snapshots = self.__simulation._metrics._temporary_snapshots

        snapshots["frame"].append(self.shared.counter)
        snapshots["id"].append(self.id)

        snapshots["x"].append(x)
        snapshots["y"].append(y)

        snapshots["image_index"].append(self._image_index)

        if self.config.image_rotation:
            angle = self.move.angle_to(Vector2((0, -1)))
            snapshots["angle"].append(round(angle))

    def __copy__(self) -> Self:
        """Create a copy of this agent and spawn it into the simulation.

        Note that this only copies the `pos` and `move` vectors.
        Any other attributes will be set to their defaults.
        """

        cls = self.__class__
        agent = cls(self._images, self.__simulation)

        # We want to make sure to copy the position and movement vectors.
        # Otherwise, the original as well as the clone will continue sharing these vectors.
        agent.pos = self.pos.copy()
        agent.move = self.move.copy()

        return agent

    def reproduce(self) -> Self:
        """Create a new agent and spawn it into the simulation.

        All values will be reset to their defaults,
        except for the agent's position and movement vector.
        These will be cloned from the original agent.
        """

        return copy(self)

    def kill(self):
        """Kill the agent.

        While violence usually isn't the right option,
        sometimes you just want to murder some innocent agents inside your little computer.

        But fear not!
        By *killing* the agent, all you're really doing is removing it from the simulation.
        """

        super().kill()

    def is_dead(self) -> bool:
        """Is the agent dead?

        Death occurs when `kill` is called.
        """
        return not self.is_alive()

    def is_alive(self) -> bool:
        """Is the agent still alive?

        Death occurs when `kill` is called.
        """

        return super().alive()



from __future__ import annotations

import random
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Type, TypeVar

import pygame as pg
from pygame.gfxdraw import hline, vline
from pygame.math import Vector2
from typing_extensions import Self

from .config import Config
from .metrics import Metrics
from .obstacle import Obstacle
from .proximity import ProximityEngine

if TYPE_CHECKING:
    from .agent import Agent

    AgentClass = TypeVar("AgentClass", bound=Agent)


@dataclass
class Shared:
    prng_move: random.Random
    """A PRNG for agent movement exclusively.
    
    To make sure that the agent's movement isn't influenced by other random function calls,
    all agents share a decoupled PRNG for movement exclusively.
    This ensures that the agents will always move the exact same way given a seed.
    """

    counter: int = 0
    """A counter that increases each tick of the simulation."""


class HeadlessSimulation:
    """The simulation class provides a snapshot of the current tick of the simulation.

    This snapshot includes the state of all agents, obstacles and sites.

    However, you don't really have to care about the internal details of the Simulation.
    Instead, you usually only want to use this class to:

    1. Create (and optionally configure) the simulation.
    2. Spawn the agents, obstacles and sites with their respective `batch_spawn_agents`, `spawn_obstacle` and `spawn_site` functions.
    3. Run the simulation!

    If a custom config isn't provided when creating the simulation, the default values of `BaseConfig` will be used instead.
    """

    shared: Shared
    """Attributes that are shared between the simulation and all agents."""

    _running: bool = False
    """The simulation keeps running as long as running is True."""

    _area: pg.rect.Rect

    # Sprite Groups
    _all: pg.sprite.Group
    _agents: pg.sprite.Group
    _obstacles: pg.sprite.Group
    _sites: pg.sprite.Group

    _next_agent_id: int = 0
    """The agent identifier to be given next."""

    # Proximity
    _proximity: ProximityEngine

    # Config that's passed on to agents as well
    config: Config
    """The config of the simulation that's shared with all agents.
    
    The config can be overriden when inheriting the Simulation class.
    However, the config must always:

    1. Inherit `Config`
    2. Be decorated by `@serde`
    """

    _metrics: Metrics
    """A collection of all the Snapshots that have been created in the simulation.
    
    Each agent produces a Snapshot at every frame in the simulation.
    """

    def __init__(self, config: Optional[Config] = None):
        self.config = config if config else Config()
        self._metrics = Metrics()

        # Initiate the seed as early as possible.
        random.seed(self.config.seed)

        # Using a custom generator for agent movement
        prng_move = random.Random()
        prng_move.seed(self.config.seed)

        self.shared = Shared(prng_move=prng_move)

        width, height = self.config.window.as_tuple()
        self._area = pg.rect.Rect(0, 0, width, height)

        # Create sprite groups
        self._all = pg.sprite.Group()
        self._agents = pg.sprite.Group()
        self._obstacles = pg.sprite.Group()
        self._sites = pg.sprite.Group()

        # Proximity!
        self._proximity = ProximityEngine(self._agents, self.config.radius)

    def batch_spawn_agents(
        self,
        count: int,
        agent_class: Type[AgentClass],
        images: list[str],
    ) -> Self:
        """Spawn multiple agents into the simulation.

        The number of agents that are spawned can be adjusted by modifying the `agent_count` option in the config.
        """

        # Load images once so the files don't have to be read multiple times.
        loaded_images = self._load_images(images)

        for _ in range(count):
            agent_class(images=loaded_images, simulation=self)

        return self

    def spawn_agent(
        self,
        agent_class: Type[AgentClass],
        images: list[str],
    ) -> Self:
        """Spawn one agent into the simulation.

        You almost always want to call `batch_spawn_agents` instead.
        This method should only be used to create a human-controllable player.
        """

        agent_class(images=self._load_images(images), simulation=self)

        return self

    def spawn_obstacle(self, image_path: str, x: int, y: int) -> Self:
        """Spawn one obstacle into the simulation. The given coordinates will be the centre of the obstacle.

        When agents collide with an obstacle, they will make a 180 degree turn.
        """

        Obstacle(
            containers=[self._all, self._obstacles],
            image=self._load_image(image_path),
            pos=Vector2((x, y)),
        )

        return self

    def spawn_site(self, image_path: str, x: int, y: int) -> Self:
        """Spawn one site into the simulation. The given coordinates will be the centre of the site."""

        Obstacle(
            containers=[self._all, self._sites],
            image=self._load_image(image_path),
            pos=Vector2((x, y)),
        )

        return self

    def run(self) -> Metrics:
        """Run the simulation until it's ended by closing the window."""

        self._running = True

        while self._running:
            self.tick()

        return self._metrics

    def before_update(self):
        """Run any code before the agents are updated in every tick.

        You should override this method when inheriting Simulation to add your own logic.

        Some examples include:
        - Processing events from PyGame's event queue.
        """

        ...

    def after_update(self):
        ...

    def tick(self):
        """Advance the simulation with one tick."""

        self.before_update()

        # Update the position of all agents
        self.__update_positions()

        # If the radius was changed by an event,
        # also update the radius in the proximity engine
        self._proximity._set_radius(self.config.radius)

        # Calculate proximity chunks
        self._proximity.update()

        # Save the replay data of all agents
        self.__collect_replay_data()

        # Update all agents
        self._all.update()

        # Merge the collected snapshots into the dataframe.
        self._metrics._merge()

        self.after_update()

        # If we've reached the duration of the simulation, then stop the simulation.
        if self.config.duration > 0 and self.shared.counter == self.config.duration:
            self.stop()
            return

        self.shared.counter += 1

    def stop(self):
        """Stop the simulation.

        The simulation isn't stopped directly.
        Instead, the current tick is completed, after which the simulation will end.
        """

        self._running = False

    def __collect_replay_data(self):
        """Collect the replay data for all agents."""

        for sprite in self._agents:
            agent: Agent = sprite  # type: ignore
            agent._collect_replay_data()

    def __update_positions(self):
        """Update the position of all agents."""

        for sprite in self._agents.sprites():
            agent: Agent = sprite  # type: ignore
            agent.change_position()

    def _load_image(self, path: str) -> pg.surface.Surface:
        return pg.image.load(path)

    def _load_images(self, images: list[str]) -> list[pg.surface.Surface]:
        return [self._load_image(path) for path in images]

    def _agent_id(self) -> int:
        agent_id = self._next_agent_id
        self._next_agent_id += 1

        return agent_id


class Simulation(HeadlessSimulation):
    _background: pg.surface.Surface
    _clock: pg.time.Clock
    _screen: pg.surface.Surface

    def __init__(self, config: Optional[Config] = None):
        super().__init__(config)

        pg.display.init()
        pg.display.set_caption("Violet")

        size = self.config.window.as_tuple()
        self._screen = pg.display.set_mode(size)

        # Initialise background
        self._background = pg.surface.Surface(size).convert()
        self._background.fill((0, 0, 0))

        # Show background immediately (before spawning agents)
        self._screen.blit(self._background, (0, 0))
        pg.display.flip()

        # Initialise the clock. Used to cap FPS.
        self._clock = pg.time.Clock()

    def before_update(self):
        rebound = []
        for event in pg.event.get(eventtype=[pg.QUIT, pg.KEYDOWN]):
            if event.type == pg.QUIT:
                self.stop()
            elif event.type == pg.KEYDOWN:
                if event.key == pg.K_HOME:
                    self.config.radius += 1
                elif event.key == pg.K_END:
                    self.config.radius -= 1
                else:
                    # If a different key was pressed, then we want to re-emit the vent
                    # so other code can handle it.
                    rebound.append(event)

        for event in rebound:
            pg.event.post(event)

        # Clear the screen before the update so agents can draw stuff themselves too!
        self._all.clear(self._screen, self._background)
        self._screen.blit(self._background, (0, 0))

    def after_update(self):
        # Draw everything to the screen
        self._all.draw(self._screen)

        if self.config.visualise_chunks:
            self.__visualise_chunks()

        # Update the screen with the new image
        pg.display.flip()

        self._clock.tick(self.config.fps_limit)

        current_fps = self._clock.get_fps()
        if current_fps > 0:
            self._metrics.fps._push(current_fps)

            if self.config.print_fps:
                print(f"FPS: {current_fps:.1f}")

    def __visualise_chunks(self):
        """Visualise the proximity chunks by drawing their borders."""

        colour = pg.Color(255, 255, 255, 122)
        chunk_size = self._proximity.chunk_size

        width, height = self.config.window.as_tuple()

        for x in range(chunk_size, width, chunk_size):
            vline(self._screen, x, 0, height, colour)

        for y in range(chunk_size, height, chunk_size):
            hline(self._screen, 0, width, y, colour)

    def _load_image(self, path: str) -> pg.surface.Surface:
        return super()._load_image(path).convert_alpha()



from __future__ import annotations

import collections
from typing import (
    TYPE_CHECKING,
    Callable,
    Generator,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from pygame.sprite import Group

if TYPE_CHECKING:
    from .agent import Agent

AgentClass = TypeVar("AgentClass", bound="Agent")
T = TypeVar("T")
U = TypeVar("U")


class ProximityIter(Generic[T]):
    """The `ProximityIter` is a small wrapper around a *generator* of agents that are in proximity.

    Now, you've probably never heard of a generator before, so let me give you the TLDR.
    A Python generator is basically a stream of values. In our case, agents!

    By not adding the agents in a list or in a set, but keeping them in a stream,
    we can add multiple filters while keeping amazing performance.

    When we're done with the filtering, we can either `count` the agents in our stream
    (thereby consuming the stream, it's single-use only) or we can collect them in a list or a set.

    Examples
    --------

    Imagine that our agent is Mudusa. We want to freeze the movement of all agents that we see.

    We can implement this by simply looping over all the agents that are returned in the `ProximityIter` stream.

    >>> class Medusa(Agent):
    ...     def update(self):
    ...         for agent, distance in self.in_proximity_accuracy():
    ...             agent.freeze_movement()

    Or perhaps we simply want to change colour if there are at least two other agents nearby.

    >>> class Chameleon(Agent):
    ...     def update(self):
    ...         if self.in_proximity_accuracy().count() >= 2:
    ...             self.change_image(1)
    ...         else:
    ...             self.change_image(0)

    In some cases, we want to loop over our stream of agents multiple times.
    To make our stream reusable, we can add its agents to a list.

    >>> class TheCollector(Agent):
    ...     def update(self):
    ...         collectables = list(self.self.in_proximity_accuracy())
    ...         # do something with our collectables multiple times!
    """

    _gen: Generator[T, None, None]

    def __init__(self, gen: Generator[T, None, None]):
        self._gen = gen

    def __iter__(self):
        return self._gen

    def filter(self, predicate: Callable[[T], bool]) -> ProximityIter[T]:
        """Filter the agents that are in proximity.

        Example
        -------

        Count the number of dead agents in proximity.

        >>> zombies = (
        ...     self.in_proximity_accuracy()
        ...     .without_distance()
        ...     .filter(lambda agent: agent.is_dead())
        ...     .count()
        ... )

        If you don't want to remove the distance,
        you can also refer to agent as the first element of the tuple:

        >>> zombies = (
        ...     self.in_proximity_accuracy()
        ...     .filter(lambda x: x[0].is_dead())
        ...     .count()
        ... )
        """

        self._gen = (agent for agent in self if predicate(agent))
        return self

    @overload
    def filter_kind(
        self: ProximityIter[tuple[AgentClass, float]], kind: Type[U]
    ) -> ProximityIter[tuple[U, float]]:
        ...

    @overload
    def filter_kind(self: ProximityIter[AgentClass], kind: Type[U]) -> ProximityIter[U]:
        ...

    def filter_kind(
        self: Union[ProximityIter[tuple[AgentClass, float]], ProximityIter[AgentClass]],
        kind: Type[U],
    ) -> Union[ProximityIter[tuple[U, float]], ProximityIter[U]]:
        """Filter the agents that are in proximity based on their class.

        Examples
        --------

        We don't want our Zombie to kill other zombies.
        Just humans!

        >>> class Human(Agent): ...

        >>> class Zombie(Agent):
        ...     def update(self):
        ...         human = (
        ...             self.in_proximity_accuracy()
        ...             .without_distance()
        ...             .filter_kind(Human)
        ...             .first()
        ...         )
        ...
        ...         if human is not None:
        ...             human.kill()
        """

        def internal_generator() -> Generator[Union[tuple[U, float], U], None, None]:
            for maybe_tuple in self:
                if isinstance(maybe_tuple, tuple):
                    agent, dist = maybe_tuple
                    if isinstance(agent, kind):
                        yield (agent, dist)
                elif isinstance(maybe_tuple, kind):
                    yield maybe_tuple

        return ProximityIter(internal_generator())  # type: ignore

    def without_distance(self: ProximityIter[tuple[U, float]]) -> ProximityIter[U]:
        """Remove the distance from the results.

        If you call `vi.agent.Agent.in_proximity_accuracy`,
        agents are returned along with their measured distance.
        However, perhaps you're not interested in the distance.

        Note that `vi.agent.Agent.in_proximity_performance` does not return the distance.
        So you cannot call this function on the performance method.

        Example
        -------

        By default, the `vi.agent.Agent.in_proximity_accuracy` method returns a stream
        of agent-distance pairs.

        >>> for agent, distance in self.in_proximity_accuracy():
        ...     # Do things with both agent and distance

        When you use `without_distance`, the distance can no longer be accessed.

        >>> for agent in self.in_proximity_accuracy().without_distance():
        ...     # Do things with agent directly
        """

        return ProximityIter(agent for agent, _ in self)

    def first(self) -> Optional[T]:
        """Retrieve the first agent that's in proximity.

        If there are no agents in proximity, `None` is returned instead.

        Examples
        --------

        Want to kill the first agent you see every frame?

        >>> other_agent = self.in_proximity_accuracy().without_distance().first()
        >>> if other_agent is not None:
        ...     other_agent.kill()

        If you don't call `without_distance`, then you cannot unpack the tuple directly,
        as it could potentially be None.
        E.g. the following code would result in a crash:

        >>> agent, distance = self.in_proximity_accuracy().first()

        Therefore, you should unpack the tuple after checking whether it is not None:

        >>> maybe_agent = self.in_proximity_accuracy().first()
        >>> if maybe_agent is not None:
        ...     agent, distance = maybe_agent
        ...     agent.kill()
        """

        return next(self._gen, None)

    def collect_set(self) -> set[T]:
        """Transform the generator into a set of agents that are in proximity.

        This is the same as wrapping the stream in a `set`.

        >>> nearby_agents = set(self.in_proximity_accuracy())

        >>> nearby_agents = self.in_proximity_accuracy().collect_set()
        """

        return set(self._gen)

    def count(self) -> int:
        """Count the number of agents that are in proximity.

        Example
        -------

        >>> in_proximity = self.in_proximity_accuracy().count()
        """

        count = 0
        for _ in self._gen:
            count += 1

        return count


class ProximityEngine:
    __agents: Group

    __chunks: dict[tuple[int, int], set[Agent]]
    """A map between chunk locations and the agents currently in that chunk."""

    chunk_size: int
    """The size of the chunks used for the proximity calculation."""

    radius: int
    """The radius representing the agent's proximity view."""

    def __init__(self, agents: Group, radius: int):
        self.__agents = agents
        self.__chunks = collections.defaultdict(set)

        self._set_radius(radius)

    def _set_radius(self, radius: int):
        self.radius = radius
        self.chunk_size = radius * 2

    def __get_chunk(self, coordinates: tuple[int, int]) -> tuple[int, int]:
        """Retrieve the chunk coordinates for an agent's coordinates."""

        x, y = coordinates

        x_chunk = x // self.chunk_size
        y_chunk = y // self.chunk_size

        return (x_chunk, y_chunk)

    def update(self):
        """Update the internal chunk store with the agents' current positions."""

        self.__chunks.clear()

        for sprite in self.__agents.sprites():
            agent: Agent = sprite  # type: ignore

            chunk = self.__get_chunk(agent.center)
            self.__chunks[chunk].add(agent)

    def __fast_retrieval(self, agent: AgentClass) -> Generator[AgentClass, None, None]:
        chunk = self.__get_chunk(agent.center)

        for nearby_agent in self.__chunks[chunk]:
            if nearby_agent.id != agent.id and agent.is_alive():
                yield nearby_agent  # type: ignore

    def in_proximity_performance(self, agent: AgentClass) -> ProximityIter[AgentClass]:
        """Retrieve a set of agents that are in the same chunk as the given agent."""

        agents = self.__fast_retrieval(agent)
        return ProximityIter(agents)

    def __accurate_retrieval(
        self, agent: AgentClass
    ) -> Generator[tuple[AgentClass, float], None, None]:
        x, y = agent.center

        CHUNK_SIZE = self.chunk_size
        RADIUS = self.radius

        x_chunk, x_offset = divmod(x, CHUNK_SIZE)
        y_chunk, y_offset = divmod(y, CHUNK_SIZE)

        x_step = 1 if x_offset >= RADIUS else -1
        x_chunk_offset = 0 if x_offset == RADIUS else x_step

        y_step = 1 if y_offset >= RADIUS else -1
        y_chunk_offset = 0 if y_offset == RADIUS else y_step

        for x in range(x_chunk, x_chunk + x_chunk_offset + x_step, x_step):
            for y in range(y_chunk, y_chunk + y_chunk_offset + y_step, y_step):
                for other in self.__chunks[(x, y)]:
                    distance = agent.pos.distance_to(other.pos)
                    if (
                        other.id != agent.id
                        and agent.is_alive()
                        and distance <= self.radius
                    ):
                        yield (other, distance)  # type: ignore

    def in_proximity_accuracy(
        self, agent: AgentClass
    ) -> ProximityIter[tuple[AgentClass, float]]:
        """Retrieve a set of agents that are in the same chunk as the given agent,
        in addition to the agents in the eight neighbouring chunks.
        """

        agents = self.__accurate_retrieval(agent)
        return ProximityIter(agents)


__all__ = ["ProximityIter"]



"""
A smol simulator framework built on top of [PyGame](https://www.pygame.org/docs/).

- Automatic agent wandering behaviour
- Fully deterministic simulations with PRNG seeds
- Install Violet with a simple `pip install` 😎
- Matrix-powered multi-threaded configuration testing
- [Polars](https://github.com/pola-rs/polars/)-powered simulation analytics
- Replay-able simulations with a ✨ time machine ✨
- Type-safe configuration system (with TOML support)

Want to get started right away?
Check out the [Violet Starter Kit](https://github.com/m-rots/violet-starter-kit)!

# A Tour of Violet

Violet is all about creating and researching collective intelligence.
And what's better than programming your own little video game to do so?

Under the hood, Violet is using [PyGame](https://www.pygame.org/docs/) to render all those fancy pixels to your very screen.
However, you don't need to have any knowledge of PyGame whatsoever to get started!

Instead, you only need to familiarise yourself with Violet's `vi.agent` and `vi.simulation` modules.
You see, all that's needed to create a *video game* is a new instance of the `vi.simulation.Simulation` class.

>>> from vi import Simulation
>>> Simulation()

Yep, that's all it takes to start your simulation.
However, it closes right away...

To actually `run` your simulation, you need to add one more thing.

>>> Simulation().run()

There, now you have a nice black window appear in the middle of your screen!

Obviously, creating a black window doesn't really count as having created a simulation just yet.
So let's add some agents!

>>> from vi import Agent, Simulation
>>> (
...     Simulation()
...     .batch_spawn_agents(100, Agent, ["examples/images/white.png"])
...     .run()
... )

We now have 100 agents wiggling around our screen.
They just don't particularly do anything just yet.
Sure, they're *moving*, but we want them to interact!

That's where you come in!
Customising the behaviour of the agents is central to creating simulations.
And it couldn't be any easier to customise what these agents do!

Violet is built around the concept of *inheritance*.
In a nutshell, inheritance allows you to well, *inherit*,
all the functions and properties that the `Agent` class exposes.
But that's not all!
Inheritance also allows you to build on top of the `Agent` class,
allowing you to implement your agent's custom behaviour simply by implementing an `update` method.

Let's see some inheritance in action.

>>> class MyAgent(Agent): ...

Here we create a new class called `MyAgent`, which inherits the `Agent` class.
Now, the three dots simply tells Python that we still have to add things to it.
But before we do that, we can already start a simulation with our new `MyAgent`.

>>> (
...     Simulation()
...     .batch_spawn_agents(100, MyAgent, ["examples/images/white.png"])
...     .run()
... )

If you look very closely, you can see that there's absolutely no difference!
The agent is still aimlessly wandering about.
And that's the key takeaway of inheritance.
Without adding or changing things,
our agent's behaviour will be exactly the same!

To actually customise our agent's behaviour,
we have to implement the `vi.agent.Agent.update` method.

The `update` method will run on every tick of the simulation,
after the positions of all the agents have been updated.
Now, you might wonder why the agent's position isn't updated in the `update` method.
Long story short, in most collective intelligence simulations,
your agent needs to interact with other agents.
Therefore, it needs to have a sense of which other agents are in its proximity.
To make sure that whenever agent A sees agent B, agent B also sees agent A,
we need to separate the position changing code from the behavioural code.
This way, either both agents see each other, or they don't see each other at all.

On the topic of agents seeing each other,
let's implement our own `update` method in which we want our agent
to change its colour whenever it sees another agent.

To be able to change colours,
we need to supply two different images to our Agent.
I'll go with a white and a red circle.

>>> (
...     Simulation()
...     .batch_spawn_agents(100, MyAgent, [
...         "examples/images/white.png",
...         "examples/images/red.png",
...     ])
...     .run()
... )

If we run the simulation again,
we see that the white image is picked automatically,
as it is the first image in the list of images.

Now, to change the currently selected image,
we can call the `vi.agent.Agent.change_image` method.

>>> class MyAgent(Agent):
...     def update(self):
...         self.change_image(1)

Remember that the first index of a Python list is 0,
so changing the image to index 1 actually selects our second image.
Cool! Our agents are now red instead! 😎

However, we don't want them to always be red.
Instead, we only want our agents to turn red when they see at least one other agent.

Fortunately, Violet already keeps track of who sees who for us, so we don't have to implement any code ourselves!
Instead, we can utilise the `vi.agent.Agent.in_proximity_accuracy` method.
This will return a `vi.proximity.ProximityIter` which we can use to count the number of agents in proximity.

Let's say that if we count at least one other agent, we turn red. Otherwise, we change the image back to white!

>>> class MyAgent(Agent):
...     def update(self):
...         if self.in_proximity_accuracy().count() >= 1:
...             self.change_image(1)
...         else:
...             self.change_image(0)

And there we have it!
A disco of a simulation with agents swapping colours whenever they get close to someone.

Now, there's way more to explore.
But this should give you an impression on how to change the behaviour of your agents with Violet.
Explore some of the modules on the left and experiment away!
"""

from dataclasses import dataclass

from pygame.math import Vector2
from serde.de import deserialize
from serde.se import serialize

from .agent import Agent
from .config import Config, Matrix, Window
from .proximity import ProximityIter
from .replay import TimeMachine
from .simulation import HeadlessSimulation, Simulation
from .util import probability



import random
from typing import Optional

from pygame.math import Vector2
from pygame.rect import Rect


def probability(threshold: float, prng: Optional[random.Random] = None) -> bool:
    """Randomly retrieve True or False depending on the given probability.

    The probability should be between 0 and 1.
    If you give a probability equal or higher than 1, this function will always return True.
    Likewise, if you give a probability equal or lower than 0, this function will always return False.
    """

    get_random = prng.random if prng else random.random
    return threshold > get_random()


def round_pos(pos: Vector2) -> tuple[int, int]:
    """Round the x and y coordinates of a vector to two integers respectively."""

    return round(pos.x), round(pos.y)


def random_angle(lenght: float, prng: Optional[random.Random] = None) -> Vector2:
    """Retrieve a randomly-angled vector with a given length."""

    uniform = prng.uniform if prng else random.uniform

    vec = Vector2(lenght, 0)
    vec.rotate_ip(uniform(0, 360))
    return vec


def random_pos(area: Rect, prng: Optional[random.Random] = None) -> Vector2:
    """Retrieve a random position within an area."""

    uniform = prng.uniform if prng else random.uniform

    x = uniform(area.left, area.right)
    y = uniform(area.top, area.bottom)
    return Vector2(x, y)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

import pygame as pg

from .config import Window

if TYPE_CHECKING:
    from polars import DataFrame, Series


def load_images(image_paths: list[str]) -> list[pg.surface.Surface]:
    return [pg.image.load(path).convert_alpha() for path in image_paths]


class TimeMachine:
    images: list[pg.surface.Surface]
    window: Window

    history: Series
    index: int = 0

    background: pg.surface.Surface
    clock: pg.time.Clock
    screen: pg.surface.Surface

    running: bool = False

    def __init__(
        self,
        history: DataFrame,
        image_paths: list[str],
        window: Optional[Window] = None,
    ):
        pg.display.init()

        # Convert multiple series (one per column) into one series of structs
        self.history = history.to_struct("agent")

        self.window = window if window is not None else Window()
        self.screen = pg.display.set_mode(self.window.as_tuple())
        pg.display.set_caption("Violet")

        # Load the images
        self.images = load_images(image_paths)

        # Initialise background
        self.background = pg.surface.Surface(self.screen.get_size()).convert()
        self.background.fill((0, 0, 0))

        # Initialise the clock. Used to cap FPS.
        self.clock = pg.time.Clock()

    def tick(self):
        for event in pg.event.get():
            if event.type == pg.QUIT:
                self.running = False
                return

        self.screen.blit(self.background, (0, 0))

        if self.index == len(self.history):
            self.running = False
            return

        current_frame: int = self.history[self.index]["frame"]

        while True:
            if self.index == len(self.history):
                self.running = False
                break

            data: dict[str, Any] = self.history[self.index]
            if data["frame"] != current_frame:
                break

            image_index = data["image_index"]
            image = self.images[image_index]

            angle = data.get("angle")
            if angle is not None:
                image = pg.transform.rotate(image, angle)

            rect = image.get_rect()
            rect.center = (data["x"], data["y"])

            self.screen.blit(image, rect)
            self.index += 1

        pg.display.flip()
        self.clock.tick(60)

    def run(self):
        self.running = True
        while self.running:
            self.tick()

        pg.quit()



import pygame as pg
from pygame.mask import Mask
from pygame.math import Vector2
from pygame.rect import Rect
from pygame.sprite import Group, Sprite
from pygame.surface import Surface

from .util import round_pos


class Obstacle(Sprite):
    image: Surface
    rect: Rect
    mask: Mask

    def __init__(self, containers: list[Group], image: Surface, pos: Vector2) -> None:
        Sprite.__init__(self, *containers)

        self.image = image
        self.rect = self.image.get_rect()
        self.rect.center = round_pos(pos)

        self.mask = pg.mask.from_surface(self.image)



from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, Generic, Optional, Type, TypeVar, Union

from serde.de import deserialize
from serde.se import serialize
from serde.toml import from_toml


def _embiggen(input_list: list[Any], copies: int):
    """The in-place deep-copy variant of list multiplication."""

    head = input_list[:]

    for _ in range(copies - 1):
        input_list.extend(deepcopy(head))


def _matrixify(matrix: dict[str, Union[Any, list[Any]]]) -> list[dict[str, Any]]:
    combinations: list[dict[str, Any]] = []

    for key, values in matrix.items():
        # Skip this key if its value is an empty list
        if isinstance(values, list) and len(values) == 0:
            continue

        # Initially the list is empty, so the dicts have to be
        # manually created on the first iteration.
        if len(combinations) == 0:
            # Multiple values
            if isinstance(values, list):
                for value in values:
                    combinations.append({key: value})

            # Single value
            elif values is not None:
                combinations.append({key: values})

        # If we have a list of dicts, we can simply add our key!
        else:
            # Multiple values
            if isinstance(values, list):
                original_length = len(combinations)
                _embiggen(combinations, len(values))

                for index, entry in enumerate(combinations):
                    value_index = index // original_length
                    entry[key] = values[value_index]

            # Single value
            elif values is not None:
                for entry in combinations:
                    entry[key] = values

    for index, entry in enumerate(combinations):
        entry["id"] = index + 1

    return combinations


@deserialize
@serialize
@dataclass
class Window:
    """Settings related to the simulation window."""

    width: int = 750
    """The width of the simulation window in pixels."""

    height: int = 750
    """The height of the simulation window in pixels."""

    @classmethod
    def square(cls, size: int):
        return cls(width=size, height=size)

    def as_tuple(self) -> tuple[int, int]:
        return (self.width, self.height)


T = TypeVar("T", bound="Config")

MatrixInt = TypeVar("MatrixInt", int, list[int])
MatrixFloat = TypeVar("MatrixFloat", float, list[float])


@deserialize
@serialize
@dataclass
class Schema(Generic[MatrixInt, MatrixFloat]):
    id: int = 0
    """The identifier of the config."""

    duration: int = 0
    """The duration of the simulation in frames.
    
    Defaults to `0`, indicating that the simulation runs indefinitely.
    """

    fps_limit: int = 60
    """Limit the number of frames-per-second.
    
    Defaults to 60 fps, equal to most screens' refresh rates.

    Set to `0` to uncap the framerate.
    """

    image_rotation: bool = False
    """Opt-in image rotation support.
    
    Please be aware that the rotation of images degrades performance by ~15%
    and currently causes a bug where agents clip into obstacles.
    """

    movement_speed: Union[float, MatrixFloat] = 0.5
    """The per-frame movement speed of the agents."""

    print_fps: bool = False
    """Print the current number of frames-per-second in the terminal"""

    radius: Union[int, MatrixInt] = 25
    """The radius (in pixels) in which agents are considered to be in proximity."""

    seed: Optional[Union[int, MatrixInt]] = None
    """The PRNG seed to use for the simulation.
    
    Defaults to `None`, indicating that no seed is used.
    """

    visualise_chunks: bool = False
    """Draw the borders of the proximity-chunks on screen."""

    window: Window = field(default_factory=Window)
    """The simulation window"""

    @classmethod
    def from_file(cls, file_name: str):
        """Load the config from a TOML file. The config doesn't have to include all attributes, only those which you want to override."""

        with open(file_name, "r") as f:
            return from_toml(cls, f.read())


@deserialize
@serialize
@dataclass
class Matrix(Schema[list[int], list[float]]):
    def to_configs(self, config: Type[T]) -> list[T]:
        """Generate a config for every unique combination of values in the matrix."""

        return [config(**values) for values in _matrixify(vars(self))]


@deserialize
@serialize
@dataclass
class Config(Schema[int, float]):
    ...


__all__ = ["Schema", "Config", "Matrix", "Window"]



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any

import polars as pl


@dataclass
class Fps:
    __fps: list[float] = field(default_factory=list)

    def _push(self, fps: float):
        self.__fps.append(fps)

    def to_polars(self) -> pl.Series:
        import polars as pl

        return pl.Series("fps", self.__fps)


class Metrics:
    """A container hosting all the accumulated simulation data over time."""

    fps: Fps
    """The frames-per-second history to analyse performance."""

    _temporary_snapshots: defaultdict[str, list[Any]]

    snapshots: pl.DataFrame
    """The [Polars DataFrame](https://pola-rs.github.io/polars-book/user-guide/quickstart/intro.html) containing the snapshot data of all agents over time."""

    def __init__(self):
        self.fps = Fps()
        self._temporary_snapshots = defaultdict(list)
        self.snapshots = pl.DataFrame()

    def _merge(self):
        df = pl.from_dict(self._temporary_snapshots)

        self.snapshots.vstack(df, in_place=True)

        self._temporary_snapshots = defaultdict(list)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vi']

package_data = \
{'': ['*']}

install_requires = \
['polars>=0.13.38,<0.14.0',
 'pygame>=2.1.2,<3.0.0',
 'pyserde[toml]>=0.7.3,<0.8.0',
 'typing-extensions>=4.2.0,<5.0.0']

setup_kwargs = {
    'name': 'violet-simulator',
    'version': '0.2.1',
    'description': 'A smol simulator framework built on top of PyGame',
    'long_description': '# Violet\n\nA smol simulator framework built on top of [PyGame](https://www.pygame.org/docs/).\n\n- Automatic agent wandering behaviour\n- Fully deterministic simulations with PRNG seeds\n- Install Violet with a simple `pip install` 😎\n- Matrix-powered multi-threaded configuration testing\n- [Polars](https://github.com/pola-rs/polars/)-powered simulation analytics\n- Replay-able simulations with a ✨ time machine ✨\n- Type-safe configuration system (with TOML support)\n\nWant to get started right away?\nCheck out the [Violet Starter Kit](https://github.com/m-rots/violet-starter-kit)!\n\n## Installation\n\nInstall the latest version of Violet with:\n\n```bash\npip3 install -U violet-simulator\n```\n\nOr with [Poetry](https://python-poetry.org):\n\n```bash\npoetry add violet-simulator\n```\n\n## Example\n\n```python\nfrom vi import Agent, Simulation\n\n(\n    # Step 1: Create a new simulation.\n    Simulation()\n    # Step 2: Add 100 agents to the simulation.\n    .batch_spawn_agents(100, Agent, images=["examples/images/white.png"])\n    # Step 3: Profit! 🎉\n    .run()\n)\n```\n\nFor more information you can check the [documentation](https://api.violet.m-rots.com), [examples](https://github.com/m-rots/violet/tree/main/examples) and the [User Guide](https://violet.m-rots.com).',
    'author': 'Storm Timmermans',
    'author_email': 'stormtimmermans@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://violet.m-rots.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
